"""
OpenTelemetry providers initialization.
Separated from config to keep configuration pure.
"""
import logging
import atexit
from typing import Optional
from opentelemetry import trace, _logs, metrics
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.semconv.resource import ResourceAttributes

try:
    from .config import (
        OTEL_ENDPOINT, OTEL_HEADERS, OTEL_PROTOCOL,
        SERVICE_NAME, SERVICE_NAMESPACE, DEPLOYMENT_ENVIRONMENT,
        RESOURCE_ATTRIBUTES,
        SPAN_EXPORT_INTERVAL_MS, LOG_EXPORT_INTERVAL_MS, METRIC_EXPORT_INTERVAL_MS,
        MAX_EXPORT_BATCH_SIZE, MAX_QUEUE_SIZE, SKIP_CLEANUP
    )
except ImportError:
    from config import (
        OTEL_ENDPOINT, OTEL_HEADERS, OTEL_PROTOCOL,
        SERVICE_NAME, SERVICE_NAMESPACE, DEPLOYMENT_ENVIRONMENT,
        RESOURCE_ATTRIBUTES,
        SPAN_EXPORT_INTERVAL_MS, LOG_EXPORT_INTERVAL_MS, METRIC_EXPORT_INTERVAL_MS,
        MAX_EXPORT_BATCH_SIZE, MAX_QUEUE_SIZE, SKIP_CLEANUP
    )

# Import console exporters for fallback
from opentelemetry.sdk.trace.export import ConsoleSpanExporter
from opentelemetry.sdk.metrics.export import ConsoleMetricExporter
from opentelemetry.sdk._logs.export import ConsoleLogExporter

# Safe console exporters that handle closed streams
class SafeConsoleSpanExporter:
    def __init__(self):
        self._exporter = ConsoleSpanExporter()
    
    def export(self, spans):
        try:
            return self._exporter.export(spans)
        except (ValueError, OSError):
            # Handle closed stdout/stderr
            return None
    
    def shutdown(self, timeout_millis=None):
        try:
            return self._exporter.shutdown()
        except (ValueError, OSError):
            return None
    
    def force_flush(self, timeout_millis=None):
        try:
            return self._exporter.force_flush(timeout_millis)
        except (ValueError, OSError):
            return None

class SafeConsoleLogExporter:
    def __init__(self):
        self._exporter = ConsoleLogExporter()
    
    def export(self, logs):
        try:
            return self._exporter.export(logs)
        except (ValueError, OSError):
            return None
    
    def shutdown(self, timeout_millis=None):
        try:
            return self._exporter.shutdown()
        except (ValueError, OSError):
            return None
    
    def force_flush(self, timeout_millis=None):
        try:
            return self._exporter.force_flush(timeout_millis)
        except (ValueError, OSError):
            return None

class SafeConsoleMetricExporter:
    def __init__(self):
        self._exporter = ConsoleMetricExporter()
        # Copy required attributes from the wrapped exporter
        self._preferred_temporality = getattr(self._exporter, '_preferred_temporality', {})
        self._preferred_aggregation = getattr(self._exporter, '_preferred_aggregation', {})
    
    def export(self, metrics_data, timeout_millis=None):
        try:
            return self._exporter.export(metrics_data, timeout_millis)
        except (ValueError, OSError):
            return None
    
    def shutdown(self, timeout=None):
        try:
            return self._exporter.shutdown()
        except (ValueError, OSError):
            return None
    
    def force_flush(self, timeout_millis=None):
        try:
            return self._exporter.force_flush(timeout_millis)
        except (ValueError, OSError):
            return None



# Choose the right exporters based on protocol
if OTEL_PROTOCOL == 'grpc':
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
    from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
    from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
else:  # http/protobuf
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
    from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter
    from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter

# Global instances - initialized once
_tracer_provider: Optional[TracerProvider] = None
_logger_provider: Optional[LoggerProvider] = None
_meter_provider: Optional[MeterProvider] = None
_tracer = None
_otel_logger = None
_logging_handler = None
_meter = None
_span_processor = None
_log_processor = None
_metric_reader = None
_atexit_registered = False

def get_resource() -> Resource:
    """Create resource with service information from config."""
    # Import here to get the latest values
    from .config import SERVICE_NAME, SERVICE_NAMESPACE, DEPLOYMENT_ENVIRONMENT, RESOURCE_ATTRIBUTES
    
    resource_dict = {
        ResourceAttributes.SERVICE_VERSION: "0.1.0",
    }
    
    # Add service info
    if SERVICE_NAME:
        resource_dict[ResourceAttributes.SERVICE_NAME] = SERVICE_NAME
    if SERVICE_NAMESPACE:
        resource_dict[ResourceAttributes.SERVICE_NAMESPACE] = SERVICE_NAMESPACE
    if DEPLOYMENT_ENVIRONMENT:
        resource_dict[ResourceAttributes.DEPLOYMENT_ENVIRONMENT] = DEPLOYMENT_ENVIRONMENT
    
    # Add any other resource attributes from config
    for key, value in RESOURCE_ATTRIBUTES.items():
        if key not in ['service.name', 'service.namespace', 'deployment.environment']:
            resource_dict[key] = value
    
    return Resource.create(resource_dict)

def get_tracer():
    """Get or create the global tracer instance."""
    global _tracer_provider, _tracer, _span_processor
    
    if _tracer is None:
        # Create resource
        resource = get_resource()
        
        # Create tracer provider
        _tracer_provider = TracerProvider(resource=resource)
        
        # Choose exporter based on configuration
        if OTEL_ENDPOINT:
            # Create OTLP exporter - add /v1/traces for traces
            trace_endpoint = OTEL_ENDPOINT
            if not trace_endpoint.endswith('/v1/traces'):
                trace_endpoint = trace_endpoint.rstrip('/') + '/v1/traces'
            
            exporter = OTLPSpanExporter(
                endpoint=trace_endpoint,
                headers=OTEL_HEADERS,
            )
        else:
            # Fallback to safe console exporter
            exporter = SafeConsoleSpanExporter()
        
        # Add batch processor
        _span_processor = BatchSpanProcessor(
            exporter,
            max_queue_size=MAX_QUEUE_SIZE,
            max_export_batch_size=MAX_EXPORT_BATCH_SIZE,
            schedule_delay_millis=SPAN_EXPORT_INTERVAL_MS,
            export_timeout_millis=5000,
        )
        
        _tracer_provider.add_span_processor(_span_processor)
        
        # Set as global tracer provider
        trace.set_tracer_provider(_tracer_provider)
        
        # Get tracer - use the actual service name from config
        from .config import SERVICE_NAME
        _tracer = trace.get_tracer(SERVICE_NAME, "0.1.0")
        
        # Register cleanup on exit
        _register_atexit()
    
    return _tracer

def get_logger_provider():
    """Get or create the global logger provider."""
    global _logger_provider, _otel_logger, _logging_handler, _log_processor
    
    if _logger_provider is None:
        # Create resource
        resource = get_resource()
        
        # Create logger provider
        _logger_provider = LoggerProvider(resource=resource)
        
        # Choose exporter based on configuration
        if OTEL_ENDPOINT:
            # Create OTLP log exporter - add /v1/logs for logs
            log_endpoint = OTEL_ENDPOINT
            if not log_endpoint.endswith('/v1/logs'):
                log_endpoint = log_endpoint.rstrip('/') + '/v1/logs'
            
            exporter = OTLPLogExporter(
                endpoint=log_endpoint,
                headers=OTEL_HEADERS,
            )
        else:
            # Fallback to safe console exporter
            exporter = SafeConsoleLogExporter()
        
        # Add batch processor
        _log_processor = BatchLogRecordProcessor(
            exporter,
            max_queue_size=MAX_QUEUE_SIZE,
            max_export_batch_size=MAX_EXPORT_BATCH_SIZE,
            schedule_delay_millis=LOG_EXPORT_INTERVAL_MS,
            export_timeout_millis=5000,
        )
        
        _logger_provider.add_log_record_processor(_log_processor)
        
        # Set as global logger provider
        _logs.set_logger_provider(_logger_provider)
        
        # Get logger
        _otel_logger = _logs.get_logger(__name__, "0.1.0")
        
        # Create logging handler
        _logging_handler = LoggingHandler(level=logging.INFO, logger_provider=_logger_provider)
        
        # Register cleanup on exit
        _register_atexit()
    
    return _logger_provider

def get_meter_provider():
    """Get or create the global meter provider."""
    global _meter_provider, _meter, _metric_reader
    
    if _meter_provider is None:
        # Create resource
        resource = get_resource()
        
        # Choose exporter based on configuration
        if OTEL_ENDPOINT:
            # Create OTLP metric exporter - add /v1/metrics for metrics
            metric_endpoint = OTEL_ENDPOINT
            if not metric_endpoint.endswith('/v1/metrics'):
                metric_endpoint = metric_endpoint.rstrip('/') + '/v1/metrics'
            
            exporter = OTLPMetricExporter(
                endpoint=metric_endpoint,
                headers=OTEL_HEADERS,
            )
        else:
            # Fallback to safe console exporter
            exporter = SafeConsoleMetricExporter()
        
        # Create metric reader
        _metric_reader = PeriodicExportingMetricReader(
            exporter=exporter,
            export_interval_millis=METRIC_EXPORT_INTERVAL_MS,
            export_timeout_millis=5000,
        )
        
        # Create meter provider
        _meter_provider = MeterProvider(
            resource=resource,
            metric_readers=[_metric_reader]
        )
        
        # Set as global meter provider
        metrics.set_meter_provider(_meter_provider)
        
        # Get meter
        _meter = metrics.get_meter(__name__, "0.1.0")
        
        # Register cleanup on exit
        _register_atexit()
    
    return _meter_provider

def get_meter():
    """Get or create the global meter instance."""
    get_meter_provider()  # Ensure provider is initialized
    return _meter

def get_logging_handler():
    """Get the OpenTelemetry logging handler."""
    get_logger_provider()  # Ensure provider is initialized
    return _logging_handler

def setup_logging(logger_name: Optional[str] = None):
    """Set up Python logging to send to OpenTelemetry."""
    handler = get_logging_handler()
    if not handler:
        return None
    
    if logger_name:
        logger = logging.getLogger(logger_name)
    else:
        logger = logging.getLogger()
    
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)
    
    return logger

def _cleanup_on_exit():
    """Cleanup function called on process exit."""
    if SKIP_CLEANUP:
        return
    
    # Production cleanup - flush first, then shutdown
    global _tracer_provider, _logger_provider, _meter_provider
    global _span_processor, _log_processor, _metric_reader
    
    try:
        # First, force flush all pending data with timeout
        if _span_processor:
            _span_processor.force_flush(timeout_millis=2000)
        if _log_processor:
            _log_processor.force_flush(timeout_millis=2000)
        if _metric_reader:
            _metric_reader.force_flush(timeout_millis=2000)
        
        # Then shutdown processors/readers
        if _span_processor:
            _span_processor.shutdown(timeout_millis=2000)
        if _log_processor:
            _log_processor.shutdown(timeout_millis=2000)
        if _metric_reader:
            _metric_reader.shutdown(timeout_millis=2000)
        
        # Finally shutdown providers
        if _tracer_provider:
            _tracer_provider.shutdown(timeout_millis=2000)
        if _logger_provider:
            _logger_provider.shutdown(timeout_millis=2000)
        if _meter_provider:
            _meter_provider.shutdown(timeout_millis=2000)
    except Exception:
        # Ignore shutdown errors to prevent crashes during exit
        pass

def _register_atexit():
    """Register the cleanup function to run on exit."""
    global _atexit_registered
    if not _atexit_registered:
        atexit.register(_cleanup_on_exit)
        _atexit_registered = True

def force_flush():
    """Force flush all telemetry data immediately (blocks)."""
    global _span_processor, _log_processor, _metric_reader
    
    if _span_processor:
        _span_processor.force_flush(timeout_millis=1000)
    if _log_processor:
        _log_processor.force_flush(timeout_millis=1000)
    if _metric_reader:
        _metric_reader.force_flush(timeout_millis=1000)

def trigger_export():
    """Just rely on the fast background export - no-op for simplicity."""
    pass

def shutdown():
    """Manually shutdown all OpenTelemetry providers."""
    _cleanup_on_exit()


"""
Test configuration and initialization functionality.
"""

import pytest
import os
from unittest.mock import patch, MagicMock


def test_parse_resource_attributes():
    """Test parsing resource attributes string."""
    from ohtell.config import parse_resource_attributes
    
    # Empty string
    assert parse_resource_attributes("") == {}
    
    # Single attribute
    result = parse_resource_attributes("service.name=test-app")
    assert result == {"service.name": "test-app"}
    
    # Multiple attributes
    result = parse_resource_attributes("service.name=test-app,service.version=1.0.0")
    assert result == {
        "service.name": "test-app",
        "service.version": "1.0.0"
    }
    
    # With spaces
    result = parse_resource_attributes("key1 = value1 , key2 = value2")
    assert result == {
        "key1": "value1",
        "key2": "value2"
    }


def test_parse_headers():
    """Test parsing headers string."""
    from ohtell.config import parse_headers
    
    # Empty string
    assert parse_headers("") == {}
    
    # Authorization header with URL encoding
    auth_header = "Authorization=Bearer%20token123"
    result = parse_headers(auth_header)
    assert result == {"Authorization": "Bearer token123"}
    
    # Multiple headers
    headers_str = "Content-Type=application/json,Accept=application/json"
    result = parse_headers(headers_str)
    assert result == {
        "Content-Type": "application/json",
        "Accept": "application/json"
    }


def test_init_function():
    """Test the init function for configuring OTEL."""
    from ohtell.config import init, SERVICE_NAME, SERVICE_NAMESPACE, DEPLOYMENT_ENVIRONMENT
    
    # Create mock config
    config = MagicMock()
    config.otel.endpoint = "http://localhost:4317"
    config.otel.headers = "Authorization=Bearer test123"
    config.otel.resource_attributes = "custom.attr=value"
    config.otel.protocol = "grpc"
    
    with patch.dict(os.environ, {}, clear=True):
        init(config, "test-app", "TestNamespace", "production")
        
        # Check environment variables were set
        assert os.environ["OTEL_EXPORTER_OTLP_ENDPOINT"] == "http://localhost:4317"
        assert os.environ["OTEL_EXPORTER_OTLP_HEADERS"] == "Authorization=Bearer test123"
        assert os.environ["OTEL_EXPORTER_OTLP_PROTOCOL"] == "grpc"
        assert os.environ["OTEL_SERVICE_NAME"] == "test-app"
        
        # Check resource attributes include our values
        resource_attrs = os.environ["OTEL_RESOURCE_ATTRIBUTES"]
        assert "service.name=test-app" in resource_attrs
        assert "service.namespace=TestNamespace" in resource_attrs
        assert "deployment.environment=production" in resource_attrs
        assert "custom.attr=value" in resource_attrs


def test_init_with_env_vars():
    """Test init function respects existing environment variables."""
    from ohtell.config import init
    
    config = MagicMock()
    config.otel = None  # No otel config
    
    with patch.dict(os.environ, {
        "OTEL_DEPLOYMENT_ENVIRONMENT": "staging",
        "ENV": "development"
    }):
        init(config, "test-app")
        
        # Should use OTEL_DEPLOYMENT_ENVIRONMENT over ENV
        resource_attrs = os.environ.get("OTEL_RESOURCE_ATTRIBUTES", "")
        assert "deployment.environment=staging" in resource_attrs


def test_init_fallback_env():
    """Test init function fallback environment detection."""
    from ohtell.config import init
    
    config = MagicMock()
    config.otel = None
    
    with patch.dict(os.environ, {"ENV": "testing"}, clear=True):
        init(config, "test-app")
        
        # Should use ENV as fallback
        resource_attrs = os.environ.get("OTEL_RESOURCE_ATTRIBUTES", "")
        assert "deployment.environment=testing" in resource_attrs


def test_setup_otel_from_config():
    """Test setup_otel_from_config function."""
    from ohtell.config import setup_otel_from_config
    
    # Config without otel section
    config_no_otel = MagicMock()
    config_no_otel.otel = None
    
    # Should not raise error
    setup_otel_from_config(config_no_otel)
    
    # Config with otel section
    config = MagicMock()
    config.otel.endpoint = "http://test:4317"
    config.otel.headers = "test-header=value"
    config.otel.resource_attributes = "test.attr=value"
    config.otel.protocol = "http/protobuf"
    
    with patch.dict(os.environ, {}, clear=True):
        setup_otel_from_config(config)
        
        assert os.environ["OTEL_EXPORTER_OTLP_ENDPOINT"] == "http://test:4317"
        assert os.environ["OTEL_EXPORTER_OTLP_HEADERS"] == "test-header=value"
        assert os.environ["OTEL_RESOURCE_ATTRIBUTES"] == "test.attr=value"
        assert os.environ["OTEL_EXPORTER_OTLP_PROTOCOL"] == "http/protobuf"


import functools
import json
import traceback
import logging
import time
from typing import Any, Optional, Dict, Callable
from datetime import datetime
import inspect

from opentelemetry import trace
from opentelemetry.trace import Status, StatusCode, SpanKind

try:
    # When installed as a package
    from .providers import get_tracer, setup_logging
    from .interceptors import PrintInterceptor
    from .metrics import get_task_metrics
except ImportError:
    # When running directly
    from providers import get_tracer, setup_logging
    from interceptors import PrintInterceptor
    from metrics import get_task_metrics


def safe_serialize(obj: Any) -> Any:
    """Safely serialize objects for span attributes."""
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj
    elif isinstance(obj, (list, tuple)):
        return [safe_serialize(item) for item in obj[:10]]  # Limit to 10 items
    elif isinstance(obj, dict):
        return {k: safe_serialize(v) for k, v in list(obj.items())[:10]}  # Limit to 10 items
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif hasattr(obj, '__dict__'):
        return f"<{obj.__class__.__name__} object>"
    else:
        return str(obj)[:100]  # Limit string representation


def add_event(event_name: str, attributes: Optional[Dict[str, Any]] = None) -> bool:
    """
    Add an event to the current active span if one exists.
    
    Args:
        event_name: Name of the event
        attributes: Optional attributes to include with the event
        
    Returns:
        True if event was added, False if no active span
    """
    current_span = trace.get_current_span()
    if current_span and current_span.is_recording():
        current_span.add_event(event_name, attributes or {})
        return True
    return False


def format_task_name(template: str, **kwargs) -> str:
    """Format task name with template variables like Prefect."""
    try:
        # Handle datetime formatting
        for key, value in kwargs.items():
            if isinstance(value, datetime) and ':' in template:
                # Extract format after colon
                pattern = f"{{{key}:([^}}]+)}}"
                import re
                match = re.search(pattern, template)
                if match:
                    date_format = match.group(1)
                    formatted_date = value.strftime(date_format.replace('%', ''))
                    template = template.replace(f"{{{key}:{match.group(1)}}}", formatted_date)
        
        # Regular string formatting for remaining variables
        return template.format(**kwargs)
    except Exception:
        return template


class task:
    """
    Decorator for tracing functions with OpenTelemetry.
    
    Similar to Prefect's @task decorator, supports:
    - Custom task names with template variables
    - Task descriptions
    - Automatic span creation and nesting
    - Input/output capture
    - Print statement interception
    - Automatic logging to SigNoz
    - Span kind differentiation (root/child)
    - Comprehensive metrics collection
    """
    
    # Class-level logger setup
    _logger_initialized = False
    _task_logger = None
    
    @classmethod
    def _ensure_logger(cls):
        """Ensure the task logger is initialized."""
        if not cls._logger_initialized:
            cls._task_logger = setup_logging("otel_wrapper.task")
            cls._logger_initialized = True
        return cls._task_logger
    
    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        task_run_name: Optional[str] = None,
        is_entrypoint: bool = False,
        trace_context_key: Optional[str] = None,
    ):
        self.name = name
        self.description = description
        self.task_run_name = task_run_name
        self.is_entrypoint = is_entrypoint
        self.trace_context_key = trace_context_key or "trace_context"
        self.tracer = get_tracer()
        self.logger = self._ensure_logger()
        self.metrics = get_task_metrics()
    
    def __call__(self, func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            # Start timing
            start_time = time.time()
            
            # Determine span name
            span_name = self.name or func.__name__
            
            # Build context for task_run_name template
            if self.task_run_name:
                # Get parameter names
                sig = inspect.signature(func)
                bound_args = sig.bind(*args, **kwargs)
                bound_args.apply_defaults()
                
                # Format the span name with parameters
                span_name = format_task_name(self.task_run_name, **bound_args.arguments)
            
            # Extract trace context if present in arguments
            trace_context_value = None
            if self.trace_context_key in kwargs:
                trace_context_value = kwargs[self.trace_context_key]
            else:
                # Look for trace_context in any of the arguments
                for arg in args:
                    if isinstance(arg, dict) and self.trace_context_key in arg:
                        trace_context_value = arg[self.trace_context_key]
                        break
                    # Also check for common trace ID patterns
                    if isinstance(arg, str) and (
                        arg.startswith('trace_') or 
                        len(arg) == 36 and arg.count('-') == 4  # UUID format
                    ):
                        trace_context_value = arg
                        break
            
            # Check if we have a current span (to determine if this is a root span)
            current_span = trace.get_current_span()
            is_root = current_span is None or not current_span.is_recording()
            
            # Determine span kind
            if self.is_entrypoint or is_root:
                span_kind = SpanKind.SERVER  # Entry point spans
            else:
                span_kind = SpanKind.INTERNAL  # Child spans
            
            # Record task start metrics
            self.metrics.record_task_start(
                task_name=span_name,
                function_name=func.__name__,
                module_name=func.__module__,
                is_root=is_root,
                is_entrypoint=self.is_entrypoint
            )
            
            # Get current trace ID for logging
            current_span = trace.get_current_span()
            current_trace_id = None
            if current_span and current_span.is_recording():
                current_trace_id = format(current_span.get_span_context().trace_id, '032x')
            
            # Log task start
            self.logger.debug(f"Task started: {span_name} (parent_trace: {current_trace_id}, is_root: {is_root})", extra={
                "task.name": span_name,
                "task.function": func.__name__,
                "task.module": func.__module__,
                "task.is_root": is_root,
                "task.is_entrypoint": self.is_entrypoint,
                "parent_trace_id": current_trace_id,
            })
            
            # Create print interceptor with logger
            print_interceptor = PrintInterceptor(logger=self.logger)
            
            # Prepare span attributes
            span_attributes = {
                "task.is_root": is_root,
                "task.is_entrypoint": self.is_entrypoint,
            }
            
            # Add trace context linking if available
            if trace_context_value:
                span_attributes["linked.trace_id"] = str(trace_context_value)
                span_attributes["trace.context"] = str(trace_context_value)
            
            # Start span with appropriate kind
            with self.tracer.start_as_current_span(
                span_name,
                kind=span_kind,
                attributes=span_attributes
            ) as span:
                # Add description as attribute
                if self.description:
                    span.set_attribute("task.description", self.description)
                
                # Add function metadata
                span.set_attribute("function.name", func.__name__)
                span.set_attribute("function.module", func.__module__)
                
                # Log the span's trace ID
                new_trace_id = format(span.get_span_context().trace_id, '032x')
                self.logger.debug(f"Span created: {span_name} (trace_id: {new_trace_id}, parent_trace: {current_trace_id})")
                
                # Add trace context linking event
                if trace_context_value:
                    span.add_event("trace_context_linked", {
                        "linked_trace_id": str(trace_context_value),
                        "span_name": span_name,
                        "function": func.__name__
                    })
                
                # Add span hierarchy info
                span.set_attribute("span.is_root", is_root)
                span.set_attribute("span.kind", span_kind.name)
                
                # Variables for metrics
                input_size = None
                output_size = None
                serialized_args = None
                serialized_kwargs = None
                serialized_result = None
                
                # Capture input arguments
                try:
                    if args:
                        serialized_args = json.dumps(safe_serialize(args))
                        span.set_attribute("function.args", serialized_args)
                        input_size = len(serialized_args.encode('utf-8'))
                        self.logger.debug(f"Task args: {serialized_args}", extra={"task.name": span_name})
                    if kwargs:
                        serialized_kwargs = json.dumps(safe_serialize(kwargs))
                        span.set_attribute("function.kwargs", serialized_kwargs)
                        kwargs_size = len(serialized_kwargs.encode('utf-8'))
                        input_size = (input_size or 0) + kwargs_size
                        self.logger.debug(f"Task kwargs: {serialized_kwargs}", extra={"task.name": span_name})
                except Exception as e:
                    span.set_attribute("function.args_error", str(e))
                
                # Execute function with print capture
                result = None
                error = None
                error_type = None
                
                try:
                    with print_interceptor.capture():
                        if inspect.iscoroutinefunction(func):
                            result = await func(*args, **kwargs)
                        else:
                            result = func(*args, **kwargs)
                    
                    # Capture output
                    try:
                        serialized_result = json.dumps(safe_serialize(result))
                        span.set_attribute("function.result", serialized_result)
                        output_size = len(serialized_result.encode('utf-8'))
                        self.logger.debug(f"Task result: {serialized_result}", extra={"task.name": span_name})
                    except Exception as e:
                        span.set_attribute("function.result_error", str(e))
                    
                    # Set success status
                    span.set_status(Status(StatusCode.OK))
                    
                    # Log task completion
                    self.logger.debug(f"Task completed: {span_name}", extra={
                        "task.name": span_name,
                        "task.status": "success",
                        "task.is_root": is_root,
                    })
                    
                except Exception as e:
                    error = e
                    error_type = type(e).__name__
                    
                    # Record exception
                    span.record_exception(e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    
                    # Add traceback
                    tb = traceback.format_exc()
                    span.set_attribute("error.traceback", tb)
                    
                    # Log error
                    self.logger.error(f"Task failed: {span_name}", extra={
                        "task.name": span_name,
                        "task.status": "error",
                        "task.is_root": is_root,
                        "error.type": error_type,
                        "error.message": str(e),
                    }, exc_info=True)
                
                finally:
                    # Calculate duration
                    end_time = time.time()
                    duration = end_time - start_time
                    
                    # Get print count
                    captured_prints = print_interceptor.get_captured_prints()
                    print_count = len(captured_prints)
                    
                    # Add captured prints as events
                    for timestamp, message in captured_prints:
                        span.add_event(
                            "print",
                            {"message": message},
                            timestamp=int(timestamp * 1e9)  # Convert to nanoseconds
                        )
                    
                    # Record task end metrics
                    self.metrics.record_task_end(
                        task_name=span_name,
                        function_name=func.__name__,
                        module_name=func.__module__,
                        is_root=is_root,
                        is_entrypoint=self.is_entrypoint,
                        duration=duration,
                        success=(error is None),
                        error_type=error_type,
                        input_size=input_size,
                        output_size=output_size,
                        print_count=print_count
                    )
                
                # Re-raise exception if occurred
                if error:
                    raise error
                
                return result
        
        # Store original function reference
        wrapper.__wrapped__ = func
        
        return wrapper


# Convenience decorators for common span types
def entrypoint(name: Optional[str] = None, description: Optional[str] = None, task_run_name: Optional[str] = None):
    """Decorator for entry point functions (e.g., API handlers, main functions)."""
    return task(name=name, description=description, task_run_name=task_run_name, is_entrypoint=True)


def traced_task(name: Optional[str] = None, description: Optional[str] = None, task_run_name: Optional[str] = None, trace_context_key: str = "trace_id"):
    """
    Decorator for tasks that automatically link trace context.
    
    Args:
        name: Task name
        description: Task description  
        task_run_name: Template for dynamic task names
        trace_context_key: Parameter name or key to look for trace context (default: "trace_id")
    
    Usage:
        @traced_task(description="Process queue message")
        def process_message(content, trace_id):
            # trace_id automatically linked in span
            pass
    """
    return task(
        name=name,
        description=description, 
        task_run_name=task_run_name,
        trace_context_key=trace_context_key
    )


def create_traced_task_with_parent(parent_trace_id: str, parent_span_id: str):
    """
    Create a task decorator that creates spans as children of the specified parent trace and span.
    
    Args:
        parent_trace_id: The parent trace ID as a hex string
        parent_span_id: The parent span ID as a hex string
    """
    def traced_task_decorator(name: str, **kwargs):
        """
        Task decorator that creates spans within the parent trace context.
        """
        def decorator(func):
            async def wrapper(*args, **func_kwargs):
                tracer = get_tracer()
                
                try:
                    # Parse the parent trace ID and span ID
                    parent_trace_id_int = int(parent_trace_id, 16)
                    parent_span_id_int = int(parent_span_id, 16)
                    
                    # Create a span context for the actual parent span
                    parent_span_context = trace.SpanContext(
                        trace_id=parent_trace_id_int,
                        span_id=parent_span_id_int,  # Use the actual parent span ID
                        is_remote=True,
                        trace_flags=trace.TraceFlags(0x01)  # SAMPLED flag
                    )
                    
                    # Create a context with the parent span
                    from opentelemetry import context
                    parent_context = trace.set_span_in_context(
                        trace.NonRecordingSpan(parent_span_context),
                        context.get_current()
                    )
                    
                    # Create a new span as a child of the parent
                    with tracer.start_as_current_span(
                        name=name,
                        context=parent_context,
                        kind=SpanKind.SERVER,
                        attributes={
                            'task.description': kwargs.get('description', ''),
                            'task.run_name': kwargs.get('task_run_name', name),
                            'parent.trace_id': parent_trace_id,
                            'parent.span_id': parent_span_id,
                            'trace.propagated': True
                        }
                    ) as span:
                        logging.debug(f"Created child span in parent trace {parent_trace_id}")
                        
                        # Set input attributes
                        span.set_attribute('function.name', func.__name__)
                        span.set_attribute('function.module', func.__module__)
                        if args or func_kwargs:
                            span.set_attribute('function.args', str(list(args) + [func_kwargs]))
                        
                        try:
                            result = await func(*args, **func_kwargs)
                            
                            # Set result attribute
                            if result is not None:
                                span.set_attribute('function.result', str(result)[:1000])  # Limit length
                            
                            span.set_status(Status(StatusCode.OK))
                            return result
                            
                        except Exception as e:
                            span.record_exception(e)
                            span.set_status(Status(StatusCode.ERROR, str(e)))
                            raise
                        
                except Exception as e:
                    # If trace context creation fails, fall back to regular task
                    logging.warning(f"Failed to create traced span with parent {parent_trace_id}:{parent_span_id}: {e}")
                    return await task(name=name, **kwargs)(func)(*args, **func_kwargs)
                    
            return wrapper
        return decorator
    return traced_task_decorator


"""
Test OTEL providers functionality.
"""

import pytest
from unittest.mock import patch, MagicMock, Mock
import logging


def test_get_tracer():
    """Test tracer creation and caching.""" 
    from ohtell.providers import get_tracer
    
    with patch('ohtell.providers.trace') as mock_trace:
        mock_tracer = MagicMock()
        mock_trace.get_tracer.return_value = mock_tracer
        
        # First call should create tracer
        tracer1 = get_tracer()
        assert tracer1 is mock_tracer
        
        # Second call should return cached tracer
        tracer2 = get_tracer()
        assert tracer2 is tracer1
        
        # get_tracer should only be called once
        mock_trace.get_tracer.assert_called_once_with("otel-wrapper", "0.1.0")


def test_get_meter():
    """Test meter creation and caching."""
    from ohtell.providers import get_meter
    
    with patch('ohtell.providers.metrics') as mock_metrics:
        mock_meter = MagicMock()
        mock_metrics.get_meter.return_value = mock_meter
        
        # First call should create meter
        meter1 = get_meter()
        assert meter1 is mock_meter
        
        # Second call should return cached meter
        meter2 = get_meter()
        assert meter2 is meter1


def test_setup_logging():
    """Test logging setup."""
    from ohtell.providers import setup_logging
    
    with patch('ohtell.providers.logging') as mock_logging:
        mock_logger = MagicMock()
        mock_logging.getLogger.return_value = mock_logger
        
        # Test with custom name
        logger = setup_logging("test_logger")
        
        mock_logging.getLogger.assert_called_with("test_logger")
        assert logger is mock_logger
        
        # Test with default name
        setup_logging()
        mock_logging.getLogger.assert_called_with("otel-wrapper")


@patch('ohtell.providers._span_processor')
@patch('ohtell.providers._log_processor') 
@patch('ohtell.providers._metric_reader')
def test_force_flush(mock_metric_reader, mock_log_processor, mock_span_processor):
    """Test force flush functionality."""
    from ohtell.providers import force_flush
    
    # Set up mocks
    mock_span_processor.force_flush.return_value = True
    mock_log_processor.force_flush.return_value = True
    mock_metric_reader.force_flush.return_value = True
    
    # Test successful flush
    result = force_flush()
    assert result is True
    
    mock_span_processor.force_flush.assert_called_once()
    mock_log_processor.force_flush.assert_called_once()  
    mock_metric_reader.force_flush.assert_called_once()


@patch('ohtell.providers.threading')
def test_trigger_export(mock_threading):
    """Test trigger export functionality."""
    from ohtell.providers import trigger_export
    
    mock_thread = MagicMock()
    mock_threading.Thread.return_value = mock_thread
    
    # Test trigger export
    trigger_export()
    
    mock_threading.Thread.assert_called_once()
    mock_thread.start.assert_called_once()


@patch('ohtell.providers._tracer_provider')
@patch('ohtell.providers._logger_provider')
@patch('ohtell.providers._meter_provider')
def test_shutdown(mock_meter_provider, mock_logger_provider, mock_tracer_provider):
    """Test shutdown functionality."""
    from ohtell.providers import shutdown
    
    # Test shutdown
    shutdown()
    
    mock_tracer_provider.shutdown.assert_called_once()
    mock_logger_provider.shutdown.assert_called_once()
    mock_meter_provider.shutdown.assert_called_once()


def test_provider_initialization():
    """Test that providers are initialized correctly."""
    # Reset global state
    import otel_wrapper.providers as providers
    providers._tracer_provider = None
    providers._logger_provider = None
    providers._meter_provider = None
    
    with patch('ohtell.providers.TracerProvider') as mock_tracer_provider_class:
        with patch('ohtell.providers.LoggerProvider') as mock_logger_provider_class:
            with patch('ohtell.providers.MeterProvider') as mock_meter_provider_class:
                with patch('ohtell.providers.Resource') as mock_resource_class:
                    with patch('ohtell.providers.OTLPSpanExporter'):
                        with patch('ohtell.providers.OTLPLogExporter'):
                            with patch('ohtell.providers.OTLPMetricExporter'):
                                
                                # Mock instances
                                mock_tracer_provider = MagicMock()
                                mock_logger_provider = MagicMock()
                                mock_meter_provider = MagicMock()
                                mock_resource = MagicMock()
                                
                                mock_tracer_provider_class.return_value = mock_tracer_provider
                                mock_logger_provider_class.return_value = mock_logger_provider
                                mock_meter_provider_class.return_value = mock_meter_provider
                                mock_resource_class.create.return_value = mock_resource
                                
                                # This should trigger initialization
                                from ohtell.providers import get_tracer
                                get_tracer()
                                
                                # Check providers were created
                                mock_tracer_provider_class.assert_called_once_with(resource=mock_resource)
                                mock_logger_provider_class.assert_called_once_with(resource=mock_resource)
                                mock_meter_provider_class.assert_called_once_with(resource=mock_resource)


import os
from pathlib import Path
from typing import Dict, Any, Optional

# Try to load config.yaml if available
try:
    from omegaconf import OmegaConf
    config_path = Path(__file__).parent.parent / "config.yaml"
    if config_path.exists():
        yaml_config = OmegaConf.load(config_path)
        otel_config = yaml_config.get('otel', {})
    else:
        otel_config = {}
except (ImportError, Exception):
    otel_config = {}

def parse_resource_attributes(attr_string: str) -> Dict[str, str]:
    """Parse resource attributes string into dictionary."""
    if not attr_string:
        return {}
    
    attrs = {}
    for pair in attr_string.split(','):
        if '=' in pair:
            key, value = pair.split('=', 1)
            attrs[key.strip()] = value.strip()
    return attrs

def parse_headers(headers_string: str) -> Dict[str, str]:
    """Parse headers string into dictionary."""
    if not headers_string:
        return {}
    
    headers = {}
    # Handle Authorization header format - decode URL encoding
    if headers_string.startswith('Authorization='):
        import urllib.parse
        auth_value = headers_string.split('=', 1)[1]
        # URL decode the value
        headers['Authorization'] = urllib.parse.unquote(auth_value)
    else:
        # Handle comma-separated key=value pairs
        for pair in headers_string.split(','):
            if '=' in pair:
                key, value = pair.split('=', 1)
                headers[key.strip()] = value.strip()
    return headers

# Configuration - Environment variables take precedence over config file
OTEL_ENDPOINT = os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT", otel_config.get('endpoint', ''))
OTEL_HEADERS_STR = os.getenv("OTEL_EXPORTER_OTLP_HEADERS", otel_config.get('headers', ''))
OTEL_PROTOCOL = os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", otel_config.get('protocol', 'http/protobuf'))
RESOURCE_ATTRIBUTES_STR = os.getenv("OTEL_RESOURCE_ATTRIBUTES", otel_config.get('resource_attributes', ''))

# Parse configuration
OTEL_HEADERS = parse_headers(OTEL_HEADERS_STR)
RESOURCE_ATTRIBUTES = parse_resource_attributes(RESOURCE_ATTRIBUTES_STR)

# Service configuration
SERVICE_NAME = os.getenv("OTEL_SERVICE_NAME", RESOURCE_ATTRIBUTES.get('service.name', 'ohtell-app'))
SERVICE_NAMESPACE = RESOURCE_ATTRIBUTES.get('service.namespace', '')
DEPLOYMENT_ENVIRONMENT = RESOURCE_ATTRIBUTES.get('deployment.environment', 'development')

# Export configuration - Environment variables with config file fallbacks
SPAN_EXPORT_INTERVAL_MS = int(os.getenv("OTEL_SPAN_EXPORT_INTERVAL_MS", otel_config.get('span_export_interval_ms', "500")))   # 0.5 seconds
LOG_EXPORT_INTERVAL_MS = int(os.getenv("OTEL_LOG_EXPORT_INTERVAL_MS", otel_config.get('log_export_interval_ms', "500")))     # 0.5 seconds  
METRIC_EXPORT_INTERVAL_MS = int(os.getenv("OTEL_METRIC_EXPORT_INTERVAL_MS", otel_config.get('metric_export_interval_ms', "30000")))  # 30 seconds

# Batch sizes - smaller for faster export
MAX_EXPORT_BATCH_SIZE = int(os.getenv("OTEL_MAX_EXPORT_BATCH_SIZE", otel_config.get('max_export_batch_size', "50")))  # Small batches
MAX_QUEUE_SIZE = int(os.getenv("OTEL_MAX_QUEUE_SIZE", otel_config.get('max_queue_size', "512")))  # Smaller queue

# Metrics sampling configuration to reduce volume
METRICS_SAMPLING_RATE = float(os.getenv("OTEL_METRICS_SAMPLING_RATE", otel_config.get('metrics_sampling_rate', "0.1")))  # Sample 10% of metrics
METRICS_ENABLED = os.getenv("OTEL_METRICS_ENABLED", str(otel_config.get('metrics_enabled', "true"))).lower() == "true"

# Cleanup configuration
SKIP_CLEANUP = os.getenv("OTEL_WRAPPER_SKIP_CLEANUP", str(otel_config.get('skip_cleanup', "true"))).lower() == "true"


def init(config: Any, app_name: str, service_namespace: str = "AI", deployment_env: Optional[str] = None) -> None:
    """
    Initialize OpenTelemetry with configuration and application name.
    
    Args:
        config: Configuration object with otel section containing:
            - endpoint: OTLP endpoint URL
            - headers: OTLP headers string
            - resource_attributes: Resource attributes string
            - protocol: OTLP protocol (optional)
        app_name: The name of the application (e.g., 'proxy-mcp-server', 'ai-core', 'ai-os-chat')
        service_namespace: The namespace group for the service (default: "AI")
        deployment_env: Deployment environment (e.g., 'production', 'staging', 'development'). 
                       If not provided, uses OTEL_DEPLOYMENT_ENVIRONMENT env var, then ENV env var, 
                       or 'development' as default.
    """
    global SERVICE_NAME, SERVICE_NAMESPACE, DEPLOYMENT_ENVIRONMENT, RESOURCE_ATTRIBUTES
    
    # Reset the tracer to ensure it uses the new service name
    from . import providers
    providers._tracer = None
    providers._tracer_provider = None
    providers._span_processor = None
    
    # Set application name and namespace
    SERVICE_NAME = app_name
    SERVICE_NAMESPACE = service_namespace
    
    # Set deployment environment - priority: function param > OTEL_DEPLOYMENT_ENVIRONMENT > ENV > default
    if deployment_env:
        DEPLOYMENT_ENVIRONMENT = deployment_env
    else:
        DEPLOYMENT_ENVIRONMENT = os.getenv("OTEL_DEPLOYMENT_ENVIRONMENT") or os.getenv("ENV", "development")
    
    # Update resource attributes
    RESOURCE_ATTRIBUTES['service.name'] = app_name
    RESOURCE_ATTRIBUTES['service.namespace'] = service_namespace
    RESOURCE_ATTRIBUTES['deployment.environment'] = DEPLOYMENT_ENVIRONMENT
    
    # Process config if provided
    if hasattr(config, 'otel') and config.otel:
        # Set environment variables for OpenTelemetry configuration
        if hasattr(config.otel, 'endpoint') and config.otel.endpoint:
            os.environ['OTEL_EXPORTER_OTLP_ENDPOINT'] = config.otel.endpoint
        
        if hasattr(config.otel, 'headers') and config.otel.headers:
            os.environ['OTEL_EXPORTER_OTLP_HEADERS'] = config.otel.headers
        
        if hasattr(config.otel, 'resource_attributes') and config.otel.resource_attributes:
            # Parse existing attributes but preserve our app_name, service_namespace, and deployment_environment
            existing_attrs = parse_resource_attributes(config.otel.resource_attributes)
            existing_attrs['service.name'] = app_name
            existing_attrs['service.namespace'] = service_namespace
            existing_attrs['deployment.environment'] = DEPLOYMENT_ENVIRONMENT
            # Update global RESOURCE_ATTRIBUTES
            RESOURCE_ATTRIBUTES.update(existing_attrs)
        
        if hasattr(config.otel, 'protocol') and config.otel.protocol:
            os.environ['OTEL_EXPORTER_OTLP_PROTOCOL'] = config.otel.protocol
    
    # Update environment variables with final resource attributes
    os.environ['OTEL_SERVICE_NAME'] = app_name
    attrs = []
    for key, value in RESOURCE_ATTRIBUTES.items():
        attrs.append(f"{key}={value}")
    os.environ['OTEL_RESOURCE_ATTRIBUTES'] = ','.join(attrs)
    
    print(f"OpenTelemetry initialized for application: {app_name}")
    print(f"  Service namespace: {service_namespace}")
    print(f"  Deployment environment: {DEPLOYMENT_ENVIRONMENT}")
    if hasattr(config, 'otel') and config.otel:
        print(f"  Endpoint: {config.otel.get('endpoint', 'Not set')}")
        print(f"  Protocol: {config.otel.get('protocol', 'http/protobuf')}")
        print(f"  Headers: {'Set' if config.otel.get('headers') else 'Not set'}")


def setup_otel_from_config(config: Any) -> None:
    """
    Setup OpenTelemetry configuration from a config object.
    
    Args:
        config: Configuration object with otel section containing:
            - endpoint: OTLP endpoint URL
            - headers: OTLP headers string
            - resource_attributes: Resource attributes string
            - protocol: OTLP protocol (optional)
    """
    if not hasattr(config, 'otel') or not config.otel:
        return
    
    # Set environment variables for OpenTelemetry configuration
    if hasattr(config.otel, 'endpoint') and config.otel.endpoint:
        os.environ['OTEL_EXPORTER_OTLP_ENDPOINT'] = config.otel.endpoint
    
    if hasattr(config.otel, 'headers') and config.otel.headers:
        os.environ['OTEL_EXPORTER_OTLP_HEADERS'] = config.otel.headers
    
    if hasattr(config.otel, 'resource_attributes') and config.otel.resource_attributes:
        os.environ['OTEL_RESOURCE_ATTRIBUTES'] = config.otel.resource_attributes
    
    if hasattr(config.otel, 'protocol') and config.otel.protocol:
        os.environ['OTEL_EXPORTER_OTLP_PROTOCOL'] = config.otel.protocol
    
    print(f"OpenTelemetry configuration set from config:")
    print(f"  Endpoint: {config.otel.get('endpoint', 'Not set')}")
    print(f"  Protocol: {config.otel.get('protocol', 'http/protobuf')}")
    print(f"  Resource attributes: {config.otel.get('resource_attributes', 'Not set')}")
    print(f"  Headers: {'Set' if config.otel.get('headers') else 'Not set'}")




"""
ohtell - OpenTelemetry Function Wrapper

A simple, decorator-based OpenTelemetry wrapper for tracing Python functions.
"""

__version__ = "0.2.6"

try:
    # When installed as a package
    from .tracer import task, entrypoint, traced_task, add_event, create_traced_task_with_parent
    from .providers import setup_logging, force_flush, trigger_export, shutdown, get_tracer, get_meter
    from .config import setup_otel_from_config, init
    from .config import *
except ImportError:
    # When running directly
    from tracer import task, entrypoint, traced_task, add_event, create_traced_task_with_parent
    from providers import setup_logging, force_flush, trigger_export, shutdown, get_tracer, get_meter
    from config import setup_otel_from_config, init
    from config import *

__all__ = [
    'task', 
    'entrypoint', 
    'traced_task', 
    'add_event', 
    'create_traced_task_with_parent', 
    'setup_logging', 
    'force_flush', 
    'trigger_export', 
    'shutdown', 
    'get_tracer', 
    'get_meter', 
    'setup_otel_from_config', 
    'init'
]